package de.ozgcloud.nachrichten.info;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.nachrichten.NachrichtenManagerProperties;
import de.ozgcloud.nachrichten.postfach.PostfachAddressIdentifier;
import de.ozgcloud.nachrichten.postfach.PostfachAddressTestFactory;
import de.ozgcloud.nachrichten.postfach.PostfachException;
import de.ozgcloud.nachrichten.postfach.PostfachMessageCode;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;

class InfoManagerServiceTest {

	@Spy
	@InjectMocks
	private InfoManagerService service;

	@Mock
	private InfoManagerRemoteService infomanagerRemoteService;
	@Mock
	private NachrichtenManagerProperties nachrichtenManagerProperties;

	@Nested
	class TestNotifyInfoManager {

		private static final String NACHRICHTEN_MANAGER_URL = "nachrichtenManagerUrl";

		private PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();

		@Test
		void shouldCallBuildInfomanagerNachricht() {
			service.notifyInfoManager(postfachNachricht);

			verify(service).builInfomanagerRequest(postfachNachricht);
		}

		@Test
		void shouldCallInfoManagerRemoteService() {
			var infomanagerNachricht = InfoManagerRequestTestFactory.create();
			doReturn(infomanagerNachricht).when(service).builInfomanagerRequest(any());

			service.notifyInfoManager(postfachNachricht);

			verify(infomanagerRemoteService).sendNotification(infomanagerNachricht);
		}

		@Test
		void shouldThrowPostfachException() {
			var causeException = new RuntimeException();
			when(infomanagerRemoteService.sendNotification(any())).thenThrow(causeException);

			var exception = assertThrows(PostfachException.class, () -> service.notifyInfoManager(postfachNachricht));
			assertThat(exception.getMessageCode()).isEqualTo(PostfachMessageCode.PROCESS_FAILED_MESSAGE_CODE);
		}

		@Nested
		class TestBuildInfomanagerNachricht {

			@Test
			void shouldSetNachrichtId() {
				var request = buildInfomanagerNachricht();

				assertThat(request.getNachrichtId()).isEqualTo(PostfachNachrichtTestFactory.ID);
			}

			@Test
			void shouldSetVorgangId() {
				var request = buildInfomanagerNachricht();

				assertThat(request.getVorgangId()).isEqualTo(PostfachNachrichtTestFactory.VORGANG_ID);
			}

			@Test
			void shouldSetPostfachId() {
				var request = buildInfomanagerNachricht();

				assertThat(request.getPostfachId()).isEqualTo(PostfachAddressTestFactory.POSTFACH_ID);
			}

			@Test
			void shouldSetNachrichtenManagerUrl() {
				when(nachrichtenManagerProperties.getAddress()).thenReturn(NACHRICHTEN_MANAGER_URL);

				var request = buildInfomanagerNachricht();

				assertThat(request.getNachrichtenManagerUrl()).isEqualTo(NACHRICHTEN_MANAGER_URL);
			}

			private InfoManagerRequest buildInfomanagerNachricht() {
				return service.builInfomanagerRequest(postfachNachricht);
			}
		}
	}

	@Nested
	class TestGetPostfachId {
		@Test
		void shouldReturnId() {
			var address = PostfachAddressTestFactory.create();

			var id = service.getPostfachId(address);

			assertThat(id).isEqualTo(PostfachAddressTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldThrowExceptionForUnkownIdentifer() {
			var identifier = new PostfachAddressIdentifier() {
			};
			var address = PostfachAddressTestFactory.createBuilder().identifier(identifier).build();

			assertThatThrownBy(() -> service.getPostfachId(address)).isInstanceOf(IllegalStateException.class);
		}
	}
}
