/*
 * Copyright (c) 2024. Das Land Schleswig-Holstein
 * vertreten durch den Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.bson.types.ObjectId;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.boot.logging.LogLevel;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.antragraum.AntragraumService;
import de.ozgcloud.nachrichten.attributes.ClientAttributeService;
import de.ozgcloud.nachrichten.info.InfoManagerService;
import de.ozgcloud.vorgang.callcontext.CallContextUserTestFactory;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import nl.altindag.log.LogCaptor;
import nl.altindag.log.model.LogEvent;

class PostfachServiceTest {

	static final String COMMAND_ID = UUID.randomUUID().toString();
	static final String USER_ID = UUID.randomUUID().toString();

	@Spy
	@InjectMocks
	private PostfachService service;

	private final PostfachRemoteService postfachRemoteService = mock(PostfachRemoteService.class);
	@Spy
	private Optional<Collection<PostfachRemoteService>> postfachRemoteServices = Optional.of(List.of(postfachRemoteService));
	@Mock
	private PersistPostfachNachrichtService persistingService;
	@Mock
	private PostfachNachrichtMapper mapper;
	@Mock
	private ApplicationEventPublisher publisher;
	@Mock
	private CurrentUserService userService;
	@Spy
	private Optional<AntragraumService> antragsraumService = Optional.of(mock(AntragraumService.class));
	@Spy
	private Optional<InfoManagerService> infomanagerService = Optional.of(mock(InfoManagerService.class));
	@Mock
	private ClientAttributeService clientAttributeService;

	@Nested
	class TestSaveDraft {

		@Captor
		private ArgumentCaptor<Optional<String>> userIdCaptor;
		@Captor
		private ArgumentCaptor<PostfachNachricht> nachrichtCaptor;

		@BeforeEach
		void init() {
			doNothing().when(service).persistPostfachNachricht(any(), any());

			when(userService.getUser()).thenReturn(CallContextUserTestFactory.create());
		}

		@Test
		void shouldUseUserIdFromContext() {
			service.saveDraft(PostfachNachrichtTestFactory.VORGANG_ID, PostfachNachrichtTestFactory.create());

			verify(service).persistPostfachNachricht(userIdCaptor.capture(), any());
			assertThat(userIdCaptor.getValue()).contains(CallContextUserTestFactory.ID);
		}

		@Test
		void shouldPersistNachricht() {
			service.saveDraft(PostfachNachrichtTestFactory.VORGANG_ID, PostfachNachrichtTestFactory.create());

			verify(service).persistPostfachNachricht(any(), any());
		}

		@Test
		void shouldAddVorgangId() {
			service.saveDraft(PostfachNachrichtTestFactory.VORGANG_ID, PostfachNachrichtTestFactory.createBuilder().vorgangId(null).build());

			verify(service).persistPostfachNachricht(any(), nachrichtCaptor.capture());
			assertThat(nachrichtCaptor.getValue().getVorgangId()).isEqualTo(PostfachNachrichtTestFactory.VORGANG_ID);
		}
	}

	@Nested
	class TestSendPostfachNachricht {

		@Mock
		private PostfachNachricht postfachNachricht;
		@Mock
		private PostfachNachricht postfachNachrichtWithSentInformation;
		@Mock
		private SendPostfachNachrichtResponse sendResponse;

		@BeforeEach
		void init() {
			doNothing().when(service).persistSentPostfachNachricht(any(), any());
			doReturn(postfachNachrichtWithSentInformation).when(service).addPostfachNachrichtSentInformation(any(), any());
			doReturn(sendResponse).when(service).handleSendPostfachNachricht(any(), any());
		}

		@Test
		void shouldCallHandleSendPostfachNachricht() {
			service.sendMail(COMMAND_ID, USER_ID, postfachNachricht);

			verify(service).handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);
		}

		@Test
		void shouldCallAddPostfachNachrichtSentInformation() {
			service.sendMail(COMMAND_ID, USER_ID, postfachNachricht);

			verify(service).addPostfachNachrichtSentInformation(postfachNachricht, sendResponse);
		}

		@Test
		void shouldCallPersistSentPostfachNachricht() {
			service.sendMail(COMMAND_ID, USER_ID, postfachNachricht);

			verify(service).persistSentPostfachNachricht(USER_ID, postfachNachrichtWithSentInformation);
		}
	}

	@Nested
	class TestAddPostfachNachrichtSentInformation {
		private static final PostfachNachricht POSTFACH_NACHRICHT = PostfachNachrichtTestFactory.createBuilder().sentAt(null).sentSuccessful(null)
				.messageCode(null).build();
		private static final SendPostfachNachrichtResponse SEND_POSTFACH_NACHRICHT_RESPONSE = SendPostfachNachrichtResponseTestFactory.create();

		@Test
		void shouldSetSentAt() {
			var result = addPostfachNachrichtSentInformation();

			assertThat(result.getSentAt()).isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
		}

		@Test
		void shouldSetSentSuccessful() {
			var result = addPostfachNachrichtSentInformation();

			assertThat(result.getSentSuccessful()).isTrue();
		}

		@Test
		void shouldSetMessageCode() {
			var result = addPostfachNachrichtSentInformation();

			assertThat(result.getMessageCode()).isEqualTo(SendPostfachNachrichtResponseTestFactory.MESSAGE_CODE.getMessageCode());
		}

		private PostfachNachricht addPostfachNachrichtSentInformation() {
			return service.addPostfachNachrichtSentInformation(POSTFACH_NACHRICHT, SEND_POSTFACH_NACHRICHT_RESPONSE);
		}
	}

	@Nested
	class TestPersistSentPostfachNachricht {

		private final PostfachNachricht outgoingPostfachNachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void init() {
			doReturn(outgoingPostfachNachricht).when(service).buildOutgoingPostfachNachricht(any(), any());
		}

		@Test
		void shouldCallBuildOutgoingPostfachNachricht() {
			var postfachNachricht = PostfachNachrichtTestFactory.create();

			service.buildOutgoingPostfachNachricht(PostfachNachrichtTestFactory.CREATED_BY, postfachNachricht);

			verify(service).buildOutgoingPostfachNachricht(PostfachNachrichtTestFactory.CREATED_BY, postfachNachricht);
		}

		@Test
		void shouldCallPersistPostfachNachricht() {
			service.persistSentPostfachNachricht(PostfachNachrichtTestFactory.CREATED_BY, PostfachNachrichtTestFactory.create());

			verify(service).persistPostfachNachricht(Optional.of(PostfachNachrichtTestFactory.CREATED_BY), outgoingPostfachNachricht);
		}
	}

	@Nested
	class TestBuildOutgoingPostfachNachricht {

		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.createBuilder().direction(null).createdAt(null)
				.createdBy(null).build();

		@Test
		void shouldSetDirectionToOut() {
			var result = buildOutgoingPostfachNachricht();

			assertThat(result.getDirection()).isEqualTo(PostfachNachricht.Direction.OUT);
		}

		@Test
		void shouldSetCreatedAt() {
			var result = buildOutgoingPostfachNachricht();

			assertThat(result.getCreatedAt()).isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
		}

		@Test
		void shouldSetCreatedBy() {
			var result = buildOutgoingPostfachNachricht();

			assertThat(result.getCreatedBy()).isEqualTo(PostfachNachrichtTestFactory.CREATED_BY);
		}

		private PostfachNachricht buildOutgoingPostfachNachricht() {
			return service.buildOutgoingPostfachNachricht(PostfachNachrichtTestFactory.CREATED_BY, postfachNachricht);
		}
	}

	@Nested
	class TestProcessForAntragsraum {

		@Nested
		class TestWithNotification {

			@BeforeEach
			void init() {
				doReturn(false).when(service).shouldNotifyAntragsraum(any());
			}

			@Test
			void shouldReturnEmpty() {
				var result = service.processForAntragsraum(PostfachNachrichtTestFactory.create());

				assertThat(result).isEmpty();
			}
		}

		@Nested
		class TestWithoutNotification {

			@BeforeEach
			void init() {
				doReturn(true).when(service).shouldNotifyAntragsraum(any());
			}

			@Test
			void shouldCallNotifyAntragsraum() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();

				service.processForAntragsraum(postfachNachricht);

				verify(infomanagerService.get()).notifyInfoManager(postfachNachricht);
			}

			@Test
			void shouldCallAdjustPostfachNachricht() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();

				service.processForAntragsraum(postfachNachricht);

				verify(service).adjustPostfachNachricht(postfachNachricht);
			}

			@Test
			void shouldReturnAdjustedPostfachNachricht() {
				var modifiedPostfachNachricht = PostfachNachrichtTestFactory.create();
				doReturn(Optional.of(modifiedPostfachNachricht)).when(service).adjustPostfachNachricht(any());

				var result = service.processForAntragsraum(PostfachNachrichtTestFactory.create());

				assertThat(result).contains(modifiedPostfachNachricht);
			}
		}

	}

	@DisplayName("Is notify antragsraum")
	@Nested
	class TestIsNotifyAntragsraum {

		@Test
		void shouldCallFindPostfachRemoteService() {
			service.shouldNotifyAntragsraum(PostfachNachrichtTestFactory.create());

			verify(service).findPostfachRemoteService(PostfachTestFactory.POSTFACH_TYPE);
		}

		@Test
		void shouldCallIsPostfachWithAntragsraum() {
			doReturn(true).when(service).isPostfachWithAntragsraum(any());
			doReturn(Optional.of(postfachRemoteService)).when(service).findPostfachRemoteService(any());

			service.shouldNotifyAntragsraum(PostfachNachrichtTestFactory.create());

			verify(service).isPostfachWithAntragsraum(postfachRemoteService);
		}

		@DisplayName("should return true if postfach with antragraum configured and")
		@ParameterizedTest(name = "replyOption is {0}")
		@EnumSource(value = PostfachNachricht.ReplyOption.class, names = { "FORBIDDEN" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldConfirm(PostfachNachricht.ReplyOption replyOption) {
			doReturn(Optional.of(postfachRemoteService)).when(service).findPostfachRemoteService(any());
			doReturn(true).when(service).isPostfachWithAntragsraum(any());
			var postfachNachricht = PostfachNachrichtTestFactory.createBuilder().replyOption(replyOption).build();

			var result = service.shouldNotifyAntragsraum(postfachNachricht);

			assertThat(result).isTrue();
		}

		@DisplayName("should return false if postfach with antragraum configured and reply is forbidden")
		@Test
		void shouldDeclineIfForbidden() {
			var postfachNachricht = PostfachNachrichtTestFactory.createBuilder().replyOption(PostfachNachricht.ReplyOption.FORBIDDEN).build();

			var result = service.shouldNotifyAntragsraum(postfachNachricht);

			assertThat(result).isFalse();
		}

		@DisplayName("should return false if no postfach is configured")
		@Test
		void shouldDeclineIfNoPostfachRemoteService() {
			doReturn(Optional.empty()).when(service).findPostfachRemoteService(any());
			var postfachNachricht = PostfachNachrichtTestFactory.createBuilder().replyOption(PostfachNachricht.ReplyOption.POSSIBLE).build();

			var result = service.shouldNotifyAntragsraum(postfachNachricht);

			assertThat(result).isFalse();
		}

		@DisplayName("should return false if configured postfach without antragsraum")
		@Test
		void shouldDeclineIfNotConfigured() {
			doReturn(Optional.of(postfachRemoteService)).when(service).findPostfachRemoteService(any());
			doReturn(false).when(service).isPostfachWithAntragsraum(any());
			var postfachNachricht = PostfachNachrichtTestFactory.createBuilder().replyOption(PostfachNachricht.ReplyOption.POSSIBLE).build();

			var result = service.shouldNotifyAntragsraum(postfachNachricht);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestAdjustPostfachNachricht {

		@Mock
		private PostfachNachricht postfachNachricht;

		@Nested
		class TestReplaceBody {

			@Test
			void shouldReturnEmptyIfNoAntragsraum() {
				ReflectionTestUtils.setField(service, "antragraumService", Optional.empty());

				var result = adjustPostfachNachricht();

				assertThat(result).isEmpty();
			}

			@Test
			void shouldAdjustMailBody() {
				when(postfachNachricht.toBuilder()).thenReturn(PostfachNachrichtTestFactory.createBuilder());
				var expectedMailBody = "Antragsraum Text";
				when(antragsraumService.get().getUserNotificationText()).thenReturn(expectedMailBody);

				var result = adjustPostfachNachricht();

				assertThat(result).isPresent().get().extracting(PostfachNachricht::getMailBody).isEqualTo(expectedMailBody);
			}

			private Optional<PostfachNachricht> adjustPostfachNachricht() {
				return service.adjustPostfachNachricht(postfachNachricht);
			}
		}
	}

	@Nested
	class TestFindById {
		@Test
		void shouldCallPersistingService() {
			service.findById(PostfachNachrichtTestFactory.ID);

			verify(persistingService).findById(PostfachNachrichtTestFactory.ID);
		}
	}

	@Nested
	class TestFindByVorgang {

		private static final String VORGANG_ID = UUID.randomUUID().toString();

		@Test
		void shouldCallPersistingService() {
			service.findByVorgang(VORGANG_ID);

			verify(persistingService).findByVorgangAsMap(VORGANG_ID);
		}
	}

	@Nested
	class TestFetchAndPersistRepliesAllPostfachs {

		@BeforeEach
		void initTest() {
			doReturn(Stream.of(postfachRemoteService)).when(service).getPostfachRemoteServices();
		}

		@Test
		void shouldCallGetPostfachRemoteServices() {
			service.fetchAndPersistReplies();

			verify(service).getPostfachRemoteServices();
		}

		@Test
		void shouldCallFetchAndPersistReplies() {
			service.fetchAndPersistReplies();

			verify(service).fetchAndPersistReplies(postfachRemoteService);
		}
	}

	@Nested
	class TestFetchAndPersistReplies {

		@Nested
		class TestHappyPath {

			private final PostfachNachricht mappedPostfachNachricht = PostfachNachrichtTestFactory.create();

			@Mock
			private PostfachNachricht postfachNachricht;

			@BeforeEach
			void init() {
				when(postfachRemoteService.getAllMessages()).thenReturn(Stream.of(postfachNachricht));
				doReturn(mappedPostfachNachricht).when(service).persistReceivedPostfachNachricht(any());
			}

			@Test
			void shouldCallGetAllMessages() {
				service.fetchAndPersistReplies(postfachRemoteService);

				verify(postfachRemoteService).getAllMessages();
			}

			@Test
			void shouldCallPersistReceivedPostfachNachricht() {
				service.fetchAndPersistReplies(postfachRemoteService);

				verify(service).persistReceivedPostfachNachricht(postfachNachricht);
			}

			@Test
			void shouldCallDeleteMessage() {
				service.fetchAndPersistReplies(postfachRemoteService);

				verify(postfachRemoteService).deleteMessage(PostfachNachrichtTestFactory.MESSAGE_ID);
			}
		}

		@Nested
		class TestWithUnsupportedOperationException {

			@Test
			void shouldHandleWhenCallGetAllMessage() {
				doThrow(new UnsupportedOperationException()).when(postfachRemoteService).getAllMessages();

				assertDoesNotThrow(() -> service.fetchAndPersistReplies(postfachRemoteService));
			}

			@Test
			void shouldHandleWhenCallDeleteMessage() {
				doThrow(new UnsupportedOperationException()).when(postfachRemoteService).deleteMessage(anyString());

				assertDoesNotThrow(() -> service.fetchAndPersistReplies(postfachRemoteService));
			}
		}
	}

	@Nested
	class TestPeristPostfachNachricht {

		@Test
		void shouldPersistPostfachNachricht() {
			var postfachNachricht = PostfachNachrichtTestFactory.create();

			service.persistPostfachNachricht(Optional.of(USER_ID), postfachNachricht);

			verify(persistingService).persistNachricht(eq(Optional.of(USER_ID)), same(postfachNachricht));
		}

	}

	@Nested
	class TestResendPostfachMail {

		private static final String COMMAND_ID = UUID.randomUUID().toString();
		private static final PostfachNachricht POSTFACH_NACHRICHT = PostfachNachrichtTestFactory.create();
		private static final Map<String, Object> RESEND_PATCH_MAP = Map.of("key", "value");

		@Mock
		private SendPostfachNachrichtResponse sendResponse;

		@BeforeEach
		void mockService() {
			when(persistingService.getById(anyString())).thenReturn(PostfachNachrichtTestFactory.asMap());
			doReturn(POSTFACH_NACHRICHT).when(mapper).fromMapToPostfachMail(anyMap());
			doReturn(sendResponse).when(service).handleSendPostfachNachricht(any(), any());
			doReturn(RESEND_PATCH_MAP).when(service).createResendPatchMap(any());
			doNothing().when(service).patchPostfachNachricht(any(), any());
		}

		@Test
		void shouldCallPersistingServiceGetById() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(persistingService).getById(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldMapFromMapToPostfachMail() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(mapper).fromMapToPostfachMail(PostfachNachrichtTestFactory.asMap());
		}

		@Test
		void shouldCallHandleSentPostfachNachricht() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(service).handleSendPostfachNachricht(COMMAND_ID, POSTFACH_NACHRICHT);
		}

		@Test
		void shouldCallCreateResendPatchMap() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(service).createResendPatchMap(sendResponse);
		}

		@Test
		void shouldCallPatchMail() {
			service.resendMail(COMMAND_ID, PostfachNachrichtTestFactory.ID);

			verify(service).patchPostfachNachricht(PostfachNachrichtTestFactory.ID, RESEND_PATCH_MAP);
		}

	}

	@Nested
	class TestCreateResendPatchMap {

		@Test
		void shouldContainsSentAt() {
			var map = service.createResendPatchMap(SendPostfachNachrichtResponseTestFactory.create());

			assertThat(map).containsKey(PostfachNachricht.FIELD_SENT_AT);
			assertThat(ZonedDateTime.parse(map.get(PostfachNachricht.FIELD_SENT_AT).toString())).isCloseTo(ZonedDateTime.now(),
					within(2, ChronoUnit.SECONDS));
		}

		@Test
		void shouldContainsSentSuccessful() {
			var map = service.createResendPatchMap(SendPostfachNachrichtResponseTestFactory.create());

			assertThat(map).containsEntry(PostfachNachricht.FIELD_SENT_SUCCESSFUL, true);
		}

		@Test
		void shouldContainsMessageCode() {
			var map = service.createResendPatchMap(SendPostfachNachrichtResponseTestFactory.create());

			assertThat(map).containsEntry(PostfachNachricht.FIELD_MESSAGE_CODE,
					SendPostfachNachrichtResponseTestFactory.MESSAGE_CODE.getMessageCode());
		}
	}

	@DisplayName("Handle send PostfachNachricht")
	@Nested
	class TestHandleSendPostfachNachricht {

		private static final String COMMAND_ID = UUID.randomUUID().toString();
		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();

		@Test
		void shouldDoSendPostfachNachricht() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

			service.handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);

			verify(service).doSendPostfachNachricht(postfachNachricht);
		}

		@Nested
		class TestSentSuccess {

			@Captor
			private ArgumentCaptor<PostfachMailSentEvent> eventCaptor;

			@Mock
			private PostfachNachricht modifiedPostfachNachricht;

			@BeforeEach
			void init() {
				doNothing().when(service).doSendPostfachNachricht(any());
			}

			@Test
			void shouldDoSendPostfachNachricht() {
				when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

				service.handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);

				verify(service).doSendPostfachNachricht(postfachNachricht);
			}

			@Test
			void shouldPublishEvent() {
				when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

				service.handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);

				verify(publisher).publishEvent(eventCaptor.capture());
				assertThat(eventCaptor.getValue().getSource()).isEqualTo(COMMAND_ID);
			}

			@Test
			void shouldReturnResponse() {
				when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

				var response = service.handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);

				assertThat(response.isSentSuccessful()).isTrue();
				assertThat(response.getMessageCode()).isEqualTo(PostfachMessageCode.SEND_SUCCESSFUL_MESSAGE_CODE);
			}

			@Test
			void shouldSendProcessedPostfachNachricht() {
				doReturn(Optional.of(modifiedPostfachNachricht)).when(service).processForAntragsraum(any());

				service.handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);

				verify(service).doSendPostfachNachricht(modifiedPostfachNachricht);
			}

		}

		@DisplayName("receive")
		@Nested
		class TestSentFailed {

			@Captor
			private ArgumentCaptor<PostfachMailSentFailedEvent> eventFailedCaptor;

			@DisplayName("OsiPostfachException")
			@Nested
			class TestOnOsiPostfachException {

				private static final String MESSAGE = "Osi Postfach throws an exception.";
				private static final PostfachException POSTFACH_EXCEPTION = new PostfachException(MESSAGE,
						PostfachMessageCode.PROCESS_FAILED_MESSAGE_CODE);

				@BeforeEach
				void mockService() {
					when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);
					doThrow(POSTFACH_EXCEPTION).when(postfachRemoteService).sendMessage(any());
				}

				@Test
				void shouldPublishEvent() {
					service.handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);

					verify(publisher).publishEvent(eventFailedCaptor.capture());
					assertThat(eventFailedCaptor.getValue().getSource()).isEqualTo(COMMAND_ID);
					assertThat(eventFailedCaptor.getValue().getErrorMessage()).startsWith(MESSAGE);
				}

				@Test
				void shouldCallLogSendPostfachNachrichtFailed() {
					service.handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);

					verify(service).logSendPostfachNachrichtFailed(COMMAND_ID, POSTFACH_EXCEPTION);
				}

				@Test
				void shouldReturnResponse() {
					var response = service.handleSendPostfachNachricht(COMMAND_ID, postfachNachricht);

					assertThat(response.isSentSuccessful()).isFalse();
					assertThat(response.getMessageCode()).isEqualTo(PostfachMessageCode.PROCESS_FAILED_MESSAGE_CODE);
				}
			}

		}
	}

	@Nested
	class TestLogSendPostfachNachrichtFailed {


		@DisplayName("should log warning when error code is unknown postfach id")
		@Test
		void shouldLogWarningWhenWrongPostfachId() {
			try (var logCaptor = LogCaptor.forClass(PostfachService.class)) {
				var exception = new PostfachException("Test", PostfachMessageCode.SEND_FAILED_UNKNOWN_POSTFACH_ID_MESSAGE_CODE);

				service.logSendPostfachNachrichtFailed(COMMAND_ID, exception);

				assertThat(logCaptor.getLogEvents()).hasSize(1).first().extracting(LogEvent::getLevel).isEqualTo(LogLevel.WARN.name());
			}
		}

		@DisplayName("should log error")
		@ParameterizedTest(name = "when error code is {0}")
		@EnumSource(value = PostfachMessageCode.class, names = { "SEND_FAILED_UNKNOWN_POSTFACH_ID_MESSAGE_CODE" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldLogError(PostfachMessageCode messageCode) {
			try (var logCaptor = LogCaptor.forClass(PostfachService.class)) {
				var exception = new PostfachException("Test", messageCode);

				service.logSendPostfachNachrichtFailed(COMMAND_ID, exception);

				assertThat(logCaptor.getLogEvents().get(0).getLevel()).isEqualTo(LogLevel.ERROR.name());
			}
		}

	}

	@Nested
	class TestDoSendPostfachNachricht {

		@Mock
		private PostfachRemoteService postfachRemoteService;

		@Test
		void shouldCallFindPostfachRemoteService() {
			doReturn(Optional.of(postfachRemoteService)).when(service).findPostfachRemoteService(any());

			service.doSendPostfachNachricht(PostfachNachrichtTestFactory.create());

			verify(service).findPostfachRemoteService(PostfachTestFactory.POSTFACH_TYPE);
		}

		@Test
		void shouldCallSendMessage() {
			doReturn(Optional.of(postfachRemoteService)).when(service).findPostfachRemoteService(any());
			var postfachNachricht = PostfachNachrichtTestFactory.create();

			service.doSendPostfachNachricht(postfachNachricht);

			verify(postfachRemoteService).sendMessage(postfachNachricht);
		}

		@Test
		void shouldThrowNotConfiguredException() {
			doReturn(Optional.empty()).when(service).findPostfachRemoteService(any());
			var postfachNachricht = PostfachNachrichtTestFactory.create();

			assertThrows(NotConfiguredException.class, () -> service.doSendPostfachNachricht(postfachNachricht));
		}
	}

	@DisplayName("Find postfach remote service")
	@Nested
	class TestFindPostfachRemoteService {

		@Mock
		private PostfachRemoteService remoteService;
		@Mock
		private PostfachRemoteService anotherRemoteService;

		private static final String MATCHING_POSTFACH_TYPE = "MatchingPostfachType";

		@Test
		void shouldReturnRemoteServiceOnMatchingPostfachType() {
			when(anotherRemoteService.getPostfachType()).thenReturn("NOT_MATCHING");
			when(remoteService.getPostfachType()).thenReturn(MATCHING_POSTFACH_TYPE);
			doReturn(Stream.of(anotherRemoteService, remoteService)).when(service).getPostfachRemoteServices();

			var foundRemoteService = service.findPostfachRemoteService(MATCHING_POSTFACH_TYPE);

			assertThat(foundRemoteService).hasValue(remoteService);
		}

		@Test
		void shouldReturnEmptyOnNotMatchingPostfachType() {
			when(remoteService.getPostfachType()).thenReturn(MATCHING_POSTFACH_TYPE);
			doReturn(Stream.of(remoteService, anotherRemoteService)).when(service).getPostfachRemoteServices();

			var foundRemoteService = service.findPostfachRemoteService("NotMatchingPostfachType");

			assertThat(foundRemoteService).isEmpty();
		}

		@DisplayName("on postfachType null")
		@Nested
		class TestOnPostfachTypeNull {

			@Test
			void shouldReturnRemoteService() {
				when(remoteService.getPostfachType()).thenReturn("BAYERN_ID");
				doReturn(Stream.of(remoteService)).when(service).getPostfachRemoteServices();

				var foundRemoteService = service.findPostfachRemoteService(null);

				assertThat(foundRemoteService).hasValue(remoteService);
			}

			@Test
			void shouldThrowExceptionMultipleConfiguredRemoteServices() {
				doReturn(Stream.of(anotherRemoteService, remoteService)).when(service).getPostfachRemoteServices();

				assertThatThrownBy(() -> service.findPostfachRemoteService(null))
						.isInstanceOf(TechnicalException.class);
			}
		}
	}

	@Nested
	class TestIsPostfachConfigured {

		@Test
		void shouldCallIsPresent() {
			service.isPostfachConfigured();

			verify(postfachRemoteServices).isPresent();
		}
	}

	@DisplayName("Get postfachs")
	@Nested
	class TestGetPostfachs {

		private final Postfach postfach = PostfachTestFactory.create();

		@Test
		void shouldCallBuildPostfach() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

			getPostfachs();

			verify(service).buildPostfach(postfachRemoteService);
		}

		@Test
		void shouldReturnPostfachs() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);
			doReturn(postfach).when(service).buildPostfach(any());

			var result = service.getPostfachs();

			assertThat(result).containsOnly(postfach);
		}

		private List<Postfach> getPostfachs() {
			return service.getPostfachs().toList();
		}
	}

	@DisplayName("on bayern id postfach")
	@Nested
	class TestOnBayernIdPostfach {

		private static final String BAYERN_ID_POSTFACH = "BAYERN_ID";
		private static final String DEPRECATED_BAYERN_ID_POSTFACH = "BayernId";

		@BeforeEach
		void mock() {
			when(postfachRemoteService.getPostfachType()).thenReturn(BAYERN_ID_POSTFACH);
		}

		@Test
		void shouldReturnTwoPostfach() {
			var postfaecher = getPostfaecher();

			assertThat(postfaecher).hasSize(2);
		}

		@Test
		void shouldContainsBayernIdCamelCase() {
			var postfaecher = getPostfaecher();

			assertThat(postfaecher).extracting(Postfach::getType).contains(DEPRECATED_BAYERN_ID_POSTFACH);
		}

		@Test
		void shouldContainsBayernIdUnderscore() {
			var postfaecher = getPostfaecher();

			assertThat(postfaecher).extracting(Postfach::getType).contains(BAYERN_ID_POSTFACH);
		}

		private Stream<Postfach> getPostfaecher() {
			return service.getPostfachs();
		}
	}

	@Nested
	class TestBuildPostfach {

		@Test
		void shouldSetPostfachType() {
			when(postfachRemoteService.getPostfachType()).thenReturn(PostfachTestFactory.POSTFACH_TYPE);

			var result = service.buildPostfach(postfachRemoteService);

			assertThat(result.getType()).isEqualTo(PostfachTestFactory.POSTFACH_TYPE);
		}

		@Test
		void shouldSetIsReplyAllowed() {
			doReturn(true).when(service).isReplyAllowed(any());

			var result = service.buildPostfach(postfachRemoteService);

			assertThat(result.isReplyAllowed()).isTrue();
		}
	}

	@Nested
	class TestIsReplyAllowed {

		@Test
		void shouldAllowWithAntragsraum() {
			doReturn(true).when(service).isPostfachWithAntragsraum(postfachRemoteService);

			var result = service.isReplyAllowed(postfachRemoteService);

			assertThat(result).isTrue();
		}

		@Test
		void shouldAllowWithPostfach() {
			doReturn(false).when(service).isPostfachWithAntragsraum(postfachRemoteService);
			when(postfachRemoteService.isReplyAllowed()).thenReturn(true);

			var result = service.isReplyAllowed(postfachRemoteService);

			assertThat(result).isTrue();
		}

		@Test
		void shouldNotAllow() {
			doReturn(false).when(service).isPostfachWithAntragsraum(postfachRemoteService);

			var result = service.isReplyAllowed(postfachRemoteService);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestPersistReceivedPostfachNachricht {

		@Test
		void shouldCallIsMukAnswer() {
			var postfachNachricht = PostfachNachrichtTestFactory.create();

			service.persistReceivedPostfachNachricht(postfachNachricht);

			verify(service).isMukAnswer(postfachNachricht);
		}

		@Nested
		class TestPostfachNachricht {

			@BeforeEach
			void init() {
				doReturn(false).when(service).isMukAnswer(any());
			}

			@Test
			void shouldCallPersistPostfachNachricht() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();

				service.persistReceivedPostfachNachricht(postfachNachricht);

				verify(service).persistPostfachNachricht(Optional.empty(), postfachNachricht);
			}

			@Test
			void shouldReturnPostfachNachricht() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();

				var result = service.persistReceivedPostfachNachricht(postfachNachricht);

				assertThat(result).isSameAs(postfachNachricht);
			}
		}

		@Nested
		class TestMukAnswer {

			private final PostfachNachricht postfachNachrichtWithVorgang = PostfachNachrichtTestFactory.create();

			@BeforeEach
			void init() {
				doReturn(true).when(service).isMukAnswer(any());
				doReturn(postfachNachrichtWithVorgang).when(service).linkToVorgang(any());
			}

			@Test
			void shouldCallLinkToVorgang() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();

				service.persistReceivedPostfachNachricht(postfachNachricht);

				verify(service).linkToVorgang(postfachNachricht);
			}

			@Test
			void shouldCallPersistPostfachNachricht() {
				service.persistReceivedPostfachNachricht(PostfachNachrichtTestFactory.create());

				verify(service).persistPostfachNachricht(Optional.empty(), postfachNachrichtWithVorgang);
			}

			@Test
			void shouldReturnUpdatedPostfachNachricht() {
				var result = service.persistReceivedPostfachNachricht(PostfachNachrichtTestFactory.create());

				assertThat(result).isSameAs(postfachNachrichtWithVorgang);
			}
		}

	}

	@Nested
	class TestSavingMukAnswer {

		private final String refId = ObjectId.get().toHexString();
		private final PostfachNachricht answer = PostfachNachrichtTestFactory.createBuilder().vorgangId(null).referencedNachricht(refId).build();

		@Test
		void shouldIdentifyAsAnswer() {
			var isMukAnswer = service.isMukAnswer(answer);

			assertThat(isMukAnswer).isTrue();
		}

		@Test
		void shouldNotIdentifyAsAnswer() {
			var isMukAnswer = service.isMukAnswer(PostfachNachrichtTestFactory.create());

			assertThat(isMukAnswer).isFalse();
		}

	}

	@Nested
	class TestLinkToVorgang {

		private static final String REFERENCE_ID = "reference-id";
		private static final PostfachNachricht POSTFACH_NACHRICHT = PostfachNachrichtTestFactory.createBuilder().vorgangId(null)
				.referencedNachricht(REFERENCE_ID).build();
		private static final PostfachNachricht PERSISTED_NACHRICHT = PostfachNachrichtTestFactory.create();
		private static final Map<String, Object> POSTFACH_NACHRICHT_MAP = PostfachNachrichtTestFactory.asMap();

		@BeforeEach
		void init() {
			when(mapper.fromMapToPostfachMail(anyMap())).thenReturn(PERSISTED_NACHRICHT);
			when(persistingService.getById(anyString())).thenReturn(POSTFACH_NACHRICHT_MAP);
		}

		@Test
		void shouldLoadPersistedNachricht() {
			service.linkToVorgang(POSTFACH_NACHRICHT);

			verify(persistingService).getById(REFERENCE_ID);
		}

		@Test
		void shouldCallFromMapToPostfachMail() {
			service.linkToVorgang(POSTFACH_NACHRICHT);

			verify(mapper).fromMapToPostfachMail(POSTFACH_NACHRICHT_MAP);
		}

		@Test
		void shouldSetVorgangId() {
			when(persistingService.getById(anyString())).thenReturn(PostfachNachrichtTestFactory.asMap());

			var result = service.linkToVorgang(POSTFACH_NACHRICHT);

			assertThat(result.getVorgangId()).isEqualTo(PostfachNachrichtTestFactory.VORGANG_ID);
		}
	}

	@Nested
	class TestPersistPostfachnachricht {

		private static final PostfachNachricht POSTFACH_NACHRICHT = PostfachNachrichtTestFactory.create();

		@Test
		void shouldCallPersistNachricht() {
			persistPostfachNachricht();

			verify(persistingService).persistNachricht(Optional.of(USER_ID), POSTFACH_NACHRICHT);
		}

		@Test
		void shouldSetClientAttribute() {
			persistPostfachNachricht();

			verify(clientAttributeService).setHasPostfachNachricht(PostfachNachrichtTestFactory.VORGANG_ID);
		}

		@Test
		@DisplayName("should NOT set hasPostfachNachricht if Nachricht is sent from system")
		void shouldNOTSetHasNachrichtForSystemNachricht() {
			var mail = PostfachNachrichtTestFactory.createBuilder().createdBy("system-123").build();

			service.persistPostfachNachricht(Optional.of(USER_ID), mail);

			verify(clientAttributeService, never()).setHasPostfachNachricht(any());
		}

		private void persistPostfachNachricht() {
			service.persistPostfachNachricht(Optional.of(USER_ID), POSTFACH_NACHRICHT);
		}
	}
}